/*
 * dxmenu-cache.c
 * 
 *  - Copyright (C) 2022 Aitor Cuadrado Zubizarreta <aitor_czr@gnuinos.org> 
 * 
 * This program is free software: you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License along
 * with this program.  If not, see <http://www.gnu.org/licenses/>.
 * 
 * See the LISENCE file.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdbool.h>
#include <libintl.h>
#include <locale.h>
#include <errno.h>
#include <fnmatch.h>
#include <signal.h>
#include <ctype.h>  // isalpha()
#include <fcntl.h>  // O_RDONLY
#include <sys/inotify.h>
#include <sys/stat.h>
#include <glib/gi18n.h>
#include <gtk/gtk.h>
#include <assert.h>
#include <sys/stat.h>
#include <pstat/libpstat.h>
#include <inttypes.h>
#include <sys/socket.h>
#include <sys/types.h>

#include "annex.h"
#include "sglib.h"

/* include config variable */
#include "config.h"

#define MAX_SIZE 1024
#define ARRAY_DIM 1
#define DATA "The menu cache file has been created"

#define strdup_or_null( str )  (str) != NULL ? strdup(str) : NULL

typedef char* cstr;

// prototypes
SGLIB_DEFINE_VECTOR_PROTOTYPES (cstr);
SGLIB_DEFINE_VECTOR_FUNCTIONS (cstr);

struct app_context {
   
   struct sglib_cstr_vector* icon;  
   struct sglib_cstr_vector* name;
   struct sglib_cstr_vector* exec;
   struct sglib_cstr_vector* category;
};

struct dir_context {
   
   struct sglib_cstr_vector* icon;  
   struct sglib_cstr_vector* name;  
   struct sglib_cstr_vector* label;  /* english name */
};

#define __cleanbuf__  __attribute__((cleanup(sbuf_free)))

typedef struct sbuf {
	char *buf;
	int bufsiz;
	int len;
} sbuf_t;

/*
 * Example life cycle:
 *	sbuf_t s;
 *	sbuf_init(&s);
 *	sbuf_addch(&s, 'F');
 *	sbuf_addstr(&s, "oo");
 *	printf("%s\n", s.buf);
 *	sbuf_free(&s);
 */

void sbuf_init(sbuf_t*);
void sbuf_addch(sbuf_t*, char);
void sbuf_addstr(sbuf_t*, const char*);
void sbuf_concat(sbuf_t*, int, ...);
void sbuf_free(sbuf_t*);

const char *progname = "dxmenu-cache";
struct dxmenu_config* config;

// Global variables
gint width, height;
int i, numDirs = 0;

struct nodeList *header = NULL;
struct nodeList *dirs = NULL;
struct nodeList *list = NULL;

gchar *cache_file = NULL;
gchar *default_terminal = NULL;
static const gchar *lang;

// Function prototypes
void get_menu_icon_path (sbuf_t*, const char*);
int display_app (struct dir_context*, struct app_context*, const char*);
int display_dir (struct dir_context*, const char*);
int display_header_item (struct app_context*, const char*);
void find_missing_icon (sbuf_t*, const char*, int, ...);
gchar* parse_cmd (const gchar*);
int update_menu_cache(struct dir_context*, struct app_context*);
int get_exec_command_from_filename (const char*, char**);
void kill_all_instances();

static
void sig_handler(int signum, siginfo_t *info, void *extra)
{ 
   switch(signum) {
         
      case SIGTERM:
      case SIGINT:
      case SIGQUIT:
      	 unlink(cache_file);
      	 g_free(cache_file);
      	 g_free(default_terminal);
         exit(EXIT_SUCCESS);  /* exit gracefully. */
         break;
         
      default:
         break;
   }
}

/*====== A subprogram like popen() but with diferences explained below ======*/
/* The pipe is to read in the current process and connected to stderr in the */
/* new one. The intent is to report error messages. Plus pid is returned.    */
/* If dirpath is an absolute path, chdir to it prior to excuting command.    */
static
FILE *epopen(const char *command,  pid_t *pid)
{
  int pipefd[2];
  pid_t p_id;
  FILE *pf;
  int rc;

  /*--------------------------- Create the pipe -----------------------------*/
  if(pipe(pipefd))
    {
      *pid = -1;
      return NULL;
    }

  /*------------------------------------ fork -------------------------------*/
  p_id = fork();
  switch(p_id)
    {
    case 0: /* the child */
      close(pipefd[0]);
      close(STDERR_FILENO);
      dup2(pipefd[1], STDERR_FILENO);
      close(pipefd[1]);
      execl( "/bin/sh", "sh", "-c", command, (char *)NULL );
      fprintf(stderr, "%s: error invoking /bin/sh: %s\n", progname, strerror(errno));
      exit(EXIT_FAILURE); /* reminder: we are the child */
    case -1: /* fork() failed ! */
      rc = errno;
      close(pipefd[0]);
      close(pipefd[1]);
      errno = rc;
      *pid = -1;
      return NULL;
    default: /* the parent */
      close(pipefd[1]);
      pf = fdopen(pipefd[0], "r");
      *pid = p_id;
      return pf; /* The caller must close the pipe after use */
    }
}

// Usage: strncpy_t(str1, sizeof(str1), str2, strlen(str2));
// Copy string "in" with at most "insz" chars to buffer "out", which
// is "outsz" bytes long. The output is always 0-terminated. Unlike
// strncpy(), strncpy_t() does not zero fill remaining space in the
// output buffer:
static
char* strncpy_t(char* out, size_t outsz, const char* in, size_t insz)
{
   assert(outsz > 0);
   while(--outsz > 0 && insz > 0 && *in) { *out++ = *in++; insz--; }
   *out = 0;
   return out;
}

// thread-safe and async-safe int to string for base 10
static
void itoa10_safe( int val, char* str ) {
   
   int rc = 0;
   int i = 0;
   int len = 0;
   int j = 0;
   bool neg = false;
   
   // sign check
   if( val < 0 ) {
      val = -val;
      neg = true;
   }
   
   // consume, lowest-order to highest-order
   if( val == 0 ) {
      str[i] = '0';
      i++;
   }
   else {
      while( val > 0 ) {
      
         int r = val % 10;
      
         str[i] = '0' + r;
         i++;
      
         val /= 10;
      }
   }
   
   if( neg ) {
      
      str[i] = '-';
      i++;
   }
   
   len = i;
   i--;
   
   // reverse order to get the number 
   while( j < i ) {
      
      char tmp = *(str + i);
      *(str + i) = *(str + j);
      *(str + j) = tmp;
      
      j++;
      i--;
   }
   
   str[len] = '\0';
}

// C substring function definition
static
char* substring(char *string, int position, int length)
{
   char *p;
   int c;

   p = malloc(length+1);
   
   if (p == NULL)
   {
      printf("Unable to allocate memory.\n");
      exit(1);
   }
 
   for (c = 0; c < length; c++)
   {
      *(p+c) = *(string+position-1);      
      string++;  
   }
 
   *(p+c) = '\0';
 
   return p;
}

static
void sglib_cstr_vector_sort (struct sglib_cstr_vector *v, int *idx, int len)
{
    int i, j;
    for (i=0; i<len; i++) idx[i]=i;
    
    for (i=0; i<len; i++)
    {
        for (j=i+1; j<len; j++)
        {
            if (strcmp(sglib_cstr_vector_at (v, idx[i]), sglib_cstr_vector_at (v, idx[j])) > 0)
            {
                int aux = idx[i];
                idx[i] = idx[j];
                idx[j] = aux;
            }
        }
    }
}

int main(int argc, char **argv)
{
    pid_t p_id, sid = 0;
	struct stat sb;
    unsigned long len;
	int rc = 0;
	//int f = -1;
	//paramid_t pi;
	int sockets[2];
	char msg[64];
    sbuf_t rc_file;
  
	/* Setting the i18n environment */
	setlocale (LC_ALL, "");
  
	/* Here, we are using -> char *bindtextdomain (const char *domainname, const char *dirname) */
	bindtextdomain ("dxmenu-cache", "/usr/share/locale");
	textdomain ("dxmenu-cache");
   
    kill_all_instances();
	
	// At this point, any previous process has been killed;
	// if argc>1, then exit doing nothing more, whatever the values of the arguments in the command line
	// We aren't parsing any argument in cache-view so far
	if(argc>1)
		return 1;

	/* process configuration and window class */
    config = (struct dxmenu_config*)calloc( sizeof(struct dxmenu_config), 1 );
    if(!config) {
      
       fprintf(stderr, "Memory allocation(): %s\n", strerror(errno));
       return -ENOMEM;
    }
   
    // config init
    rc = dxmenu_config_init( config );
    if( rc != 0 ) {
   
      
      return rc;
    }

    // if we didn't get a config file, use the default one
    if( config->config_path == NULL ) {
      
        sbuf_init(&rc_file);
        sbuf_concat(&rc_file, 2, getenv("HOME"), "/.dxmenurc");
        if(access(rc_file.buf, F_OK) == 0)          
            config->config_path = strdup_or_null(rc_file.buf);
        else
            config->config_path = strdup_or_null("/etc/default/dxmenurc");
      
        if( config->config_path == NULL ) {

            // OOM 
            return -ENOMEM;
        }
    }
      
  sbuf_free(&rc_file);
   
   // load from file...
   rc = dxmenu_config_load( config->config_path, config );
   if( rc != 0 ) {
	   
      fprintf(stderr, "dxmenu_config_load('%s') rc = %d\n", config->config_path, rc );
      return rc;
   }
 	
    cache_file = g_strconcat(getenv("HOME"), "/.cache/dxmenu-cache", NULL);
   
	if (socketpair(AF_UNIX, SOCK_STREAM, 0, sockets) < 0) {
		perror("opening stream socket pair");
		exit(1);
	}
	
    p_id = fork();
   
	if( p_id == 0 ) {   /* child */
	
		struct sigaction sa;
		char *buf = NULL;
		int *idx;
        struct sglib_cstr_vector icon;  
        struct sglib_cstr_vector name;
        struct sglib_cstr_vector exec;
        struct sglib_cstr_vector category;
        struct sglib_cstr_vector dir_icon;  
        struct sglib_cstr_vector dir_name;  
        struct sglib_cstr_vector dir_label;
        struct app_context app_ctx;
        struct dir_context dir_ctx;
		
		int rc = 0;
		int inotfd = -1;
		int watch_usr = -1;
		int watch_usrlocal = -1; 
		int watch_xdg = -1;
		fd_set rfds;
      
        close(sockets[1]);
	
		memset (&sa, 0, sizeof(sa));
     
		/*---------------- initialize signal handling ------------ */
		sa.sa_sigaction = &sig_handler;
		sa.sa_flags = SA_RESETHAND | SA_NODEFER | SA_SIGINFO | SA_RESTART;
		sigemptyset(&sa.sa_mask);
		sigaction(SIGTERM, &sa, 0);
		sigaction(SIGINT,  &sa, 0);
		sigaction(SIGQUIT, &sa, 0);
		
		// We do need gtk in order get the icon images according to the current theme 
		if (!gtk_init_check (NULL, NULL)) {
			g_print("Error, cannot initialize gtk.\n");
			exit (EXIT_FAILURE);
		}

		// Global system variables
		gchar l[16] = {0};
		g_stpcpy(l, getenv("LANG"));
		lang = strtok(l, "_");

		g_unsetenv("XDG_MENU_PREFIX");

		gtk_init (NULL, NULL);

		if (!gtk_icon_size_lookup (GTK_ICON_SIZE_BUTTON, &width, &height)) { 
			width = 24;
			height = 24;
		};
  
        sglib_cstr_vector_init (&icon);
        sglib_cstr_vector_init (&name);
        sglib_cstr_vector_init (&exec);
        sglib_cstr_vector_init (&category);
    
        sglib_cstr_vector_init (&dir_icon);
        sglib_cstr_vector_init (&dir_name);
        sglib_cstr_vector_init (&dir_label);
  
        memset (&app_ctx, 0, sizeof(struct app_context));
        memset (&app_ctx, 0, sizeof(struct dir_context));
  
        app_ctx.icon = &icon;
        app_ctx.name = &name;
        app_ctx.exec = &exec;
        app_ctx.category = &category;
    
        dir_ctx.icon = &dir_icon;
        dir_ctx.name = &dir_name;
        dir_ctx.label = &dir_label;

		inotfd = inotify_init1(IN_CLOEXEC);
		if (inotfd < 0) {
			printf( "cannot initialize inotify: %s\n", strerror(errno) );
			exit(EXIT_FAILURE);
		}
    
		watch_usr = inotify_add_watch(inotfd, "/usr/share/applications", IN_CREATE | IN_DELETE | IN_MODIFY);
		if(watch_usr == -1) {

			printf("cannot add inotify watch '/usr/share/applications': %s\n", strerror(errno));
			exit(EXIT_FAILURE);
		}
    
		watch_usrlocal = inotify_add_watch(inotfd, "/usr/local/share/applications", IN_CREATE | IN_DELETE | IN_MODIFY);
		if(watch_usrlocal == -1) {
	      
			// do not exit here
			//printf("cannot add inotify watch '/usr/local/share/applications': %s\n", strerror(errno));
		}
    /*
		watch_xdg = inotify_add_watch(inotfd, "/etc/xdg/menus", IN_CREATE | IN_DELETE | IN_MODIFY);
		if(watch_xdg == -1) {

			printf("cannot add inotify watch '/etc/xdg/menus': %s\n", strerror(errno));
			exit(EXIT_FAILURE);
		}
     */

			
		FILE *f = fopen(cache_file, "w");
		if(!f) {
		    printf("file can't be opened\n");
			exit(EXIT_FAILURE);
		}
			
        rc = display_header_item(&app_ctx, config->terminal_emulator);
        if(rc < 0) goto Sglib_free;
  
        rc = display_header_item(&app_ctx, config->file_manager);
        if(rc < 0) goto Sglib_free;
   
        rc = display_header_item(&app_ctx, config->web_browser);
        if(rc < 0) goto Sglib_free;
  
        rc = display_header_item(&app_ctx, config->mail_client);
        if(rc < 0) goto Sglib_free;

        rc = get_exec_command_from_filename (config->application_launcher, &buf);
        if(rc == 0) {
	        
	        sbuf_t s;
	        sbuf_init(&s);
	        get_menu_icon_path (&s, "system-run");   
            fprintf(f, "IMG:%s	%s	%s\n\n", s.buf, _("execute"), buf);
            sbuf_free(&s);
        }
 
        free(buf);
        fflush(stdout);
            
        get_exec_command_from_filename (config->terminal_emulator, &buf);
        if(rc == 0)            
            default_terminal = strdup_or_null(buf);
        else            
            default_terminal = strdup_or_null("xterm");
        free(buf);
                                  
        rc = update_menu_cache(&dir_ctx, &app_ctx);
        if(rc < 0) goto Sglib_free;
        
        len = sglib_cstr_vector_size (&name);
 
        i = 0;
        while(i<len) {   
	
	        const char *category_str = sglib_cstr_vector_at (app_ctx.category, i);
	        if (strcmp(category_str, "menu_header")) break;
	        fprintf(f, "IMG:%s	%s	%s\n", sglib_cstr_vector_at (app_ctx.icon, i), sglib_cstr_vector_at (app_ctx.name, i), parse_cmd(sglib_cstr_vector_at (app_ctx.exec, i)));
	        i++;
        }
        fprintf(f, "\n");
  
        len = sglib_cstr_vector_size (&dir_name);
		
	    idx = (int*)malloc(sizeof(int) * len);
	    if (!idx) {
		
		    fprintf(stderr, "Memory allocation failure\n");
		    exit (EXIT_FAILURE);
	    }
  
        sglib_cstr_vector_sort (&dir_name, idx, len);
   
        i = 0;
        while(i<len) {
	    
	        int j = 0, k = 0;
	        unsigned long l = sglib_cstr_vector_size (&name); 
	        const char *str = sglib_cstr_vector_at (dir_ctx.name, idx[i]);
	        
	        while(j<l) {
		        
		        const char *category_str = sglib_cstr_vector_at (app_ctx.category, j);
		        if (category_str && !strcmp(category_str, str)) {
			    
			        if(k == 0)			  
			            fprintf(f, "IMG:%s	%s\n", sglib_cstr_vector_at (dir_ctx.icon, idx[i]), str);
		            k++;
			        fprintf(f, "	IMG:%s	%s	%s\n", sglib_cstr_vector_at (app_ctx.icon, j), sglib_cstr_vector_at (app_ctx.name, j), parse_cmd(sglib_cstr_vector_at (app_ctx.exec, j)));
		        }
	 	        j++;
	        }
	        i++;
        }

        rc = get_exec_command_from_filename (config->logout_dialog, &buf);
        if(rc == 0) {
	    
	        sbuf_t s;
	        sbuf_init(&s);
	        get_menu_icon_path (&s, "system-shutdown");
            fprintf(f, "\nIMG:%s	%s	%s", s.buf, _("exit"), buf);
            sbuf_free(&s);
        }

        free(buf);
	    fclose(f);
    
Sglib_free:

        free(idx); 

        sglib_cstr_vector_free (&icon);
        sglib_cstr_vector_free (&name);
        sglib_cstr_vector_free (&exec);
        sglib_cstr_vector_free (&category);
  
        sglib_cstr_vector_free (&dir_icon);
        sglib_cstr_vector_free (&dir_name);
        sglib_cstr_vector_free (&dir_label);

        if (write(sockets[0], DATA, sizeof(DATA)) < 0) {
			
            perror("writing stream message");
        }    
      
        close(sockets[0]);

Select:			
		FD_ZERO(&rfds);
		FD_SET(inotfd, &rfds);
		
		rc = select(inotfd + 1, &rfds, NULL, NULL, NULL);
		if (rc < 0) {

			if (errno == EINTR) {
				rc = -errno;
				printf("select(): '%s' errno = %d\n", strerror(errno), rc);
				goto Select;
			}
		}
   
		// reset the inotify watch descriptors 
		if( watch_usr != -1 ) {
				
			rc = inotify_rm_watch( inotfd, watch_usr );
			if( rc < 0 ) {
				rc = -errno;
				printf("close(events_wd %d) rc = %d", watch_usr, rc );
			}
			else {
              
				watch_usr = -1;
			}
		}
   
		if( watch_usrlocal != -1 ) {
	    
			rc = inotify_rm_watch( inotfd, watch_usrlocal );
			if( rc < 0 ) {

         
				rc = -errno;
				printf("close(events_wd %d) rc = %d", watch_usrlocal, rc );
			}
			else {
         
				watch_usrlocal = -1;
			}
		}
		
		if( watch_xdg != -1 ) {
				
			rc = inotify_rm_watch( inotfd, watch_xdg );
			if( rc < 0 ) {
				rc = -errno;
				printf("close(events_wd %d) rc = %d", watch_xdg, rc );
			}
			else {
              
				watch_xdg = -1;
			}
		}		

		/* FD_ISSET(0, &rfds) will be true. */
		if ( FD_ISSET(inotfd, &rfds) ) {
   
     		if( access( cache_file, F_OK ) == 0 ) {
      	 
				remove(cache_file);
				g_free(cache_file);
				g_free(default_terminal);
				sleep(1);
				char *const params[] = {"dxmenu-cache", (char*)0};
				execv("/usr/bin/dxmenu-cache", params);
			}
			else {
				g_free(cache_file);
				g_free(default_terminal);
			}
		}	  
	}
	else if(p_id == -1) {  /* fork() failed ! */
      
       rc = errno;
       
       
       errno = rc;
       exit(EXIT_FAILURE);
    }
    else { /* the parent */
      
       close(sockets[0]);
       if (read(sockets[1], msg, sizeof(msg)) < 0)
         perror("reading stream message");
         
       printf("%s\n", msg);
      
       close(sockets[1]);
       
       return 0;
    }
}

int update_menu_cache(struct dir_context *dir_ctx, struct app_context *app_ctx)
{
    int rc = 0;
    DIR *pDir = NULL;
    DIR *pApp = NULL;
    struct dirent *dir_ent = NULL;
    struct dirent *app_ent = NULL;
    
    pDir = opendir("/usr/share/desktop-directories");
    while ((dir_ent = readdir(pDir))) {
		
        if(strcmp(dir_ent->d_name, ".") && strcmp(dir_ent->d_name, "..")) {
			
			int m = 0;
			sbuf_t path;
			char *category = NULL;
			//const char * whitelist[] = { "Utility", "Graphics", "Education", "Network", "Game", "Office", "Development", "AudioVideo", "System", "Settings", (char*)0 };
			const char * params_1[] = { "utility", "graphics", "education", "network", "game", "office", "development", "audio-video", "system", "settings", (char*)0 };
			const char * params_2[] = { "Utility", "Graphics", "Education", "Network", "Game", "Office", "Development", "AudioVideo", "System", "Settings", (char*)0 };
			
			sbuf_init(&path);
			sbuf_concat(&path, 2, "/usr/share/desktop-directories/", dir_ent->d_name);
            char *basename = rindex(path.buf, '/') + 1;
            char *extension = rindex(basename, '.');
            if (strcmp(extension, ".directory")) {
				
                sbuf_free(&path);
                continue;
            }
            
            char *prefix = substring(basename, 1, 5);
            
            if (strcmp(prefix, "lxde-")) {
				
                sbuf_free(&path);
                free(prefix);
                continue;
            }  
    
            category = substring(basename, 6, strlen(basename)-15);
					
			while(params_1[m]) {
               if (!strcmp(category, params_1[m])) {
				  rc = display_dir(dir_ctx, path.buf);
                  if(rc < 0) {
                     sbuf_free(&path);
                     free(prefix);
				     closedir(pDir);
				     return rc;
				  }

                  rc = sglib_cstr_vector_push_back (dir_ctx->label, strdup_or_null(params_2[m]));
                  if(rc < 0) {
                     sbuf_free(&path);
                     free(prefix);
				     closedir(pDir);
				     return rc;
				  }
				  
				  break;
			   }
			   
			   m++;  
            }
               
            sbuf_free(&path);
            free(prefix);
        }
    }    
    
    closedir(pDir);
    
    pApp = opendir("/usr/share/applications");
    while ((app_ent = readdir(pApp))) {
        if(strcmp(app_ent->d_name, ".") && strcmp(app_ent->d_name, "..")) {
			sbuf_t path;
			sbuf_init(&path);
			sbuf_concat(&path, 2, "/usr/share/applications/", app_ent->d_name);
            char *tmp = rindex(path.buf, '.');
            if (!tmp || strcmp(tmp, ".desktop")) {
                sbuf_free(&path);
                continue;
            }
            rc = display_app(dir_ctx, app_ctx, path.buf);
            if(rc < 0) {
               sbuf_free(&path);
               closedir(pApp);
               return rc;
			}
            sbuf_free(&path);
        }
    }    
        
    closedir(pApp);
    return rc;    
}

int display_app (struct dir_context *dir_ctx, struct app_context *app_ctx, const char *path)
{
    int rc = 0;
    char buffer[MAX_SIZE] = {0};
    char *eol;
    sbuf_t name, exec, icon, category, entry, type;
    FILE *fp = NULL;
    bool display_ok = true;
    bool show_item_ok = true;
    bool terminal_ok = false;
   
    fp = fopen(path, "r");
    if(!fp) {
        exit(EXIT_FAILURE);
    }
    
    eol = malloc(MAX_SIZE * sizeof(char));  
    if(!eol) {
        printf("Memory failure\n");
        exit( EXIT_FAILURE );
    }
    
    while(eol[0] != '[') fgets(eol, MAX_SIZE, fp);

    sbuf_init(&entry);
    if(lang && !strcmp(lang, "en"))
        sbuf_addstr(&entry, "Name");
    else
        sbuf_concat(&entry, 4, "Name", "[", lang, "]");
        
    while(fgets(eol, MAX_SIZE, fp)>0) {
        eol[strcspn(eol, "\n")] = 0;
        if(eol[0] == '[') break;
        else if(eol[0] == '#' || isalpha(eol[0]) == 0) continue; 
        else {
            
            char *rest = eol;
            char *key = strtok_r(rest, "=", &rest);
            
            if (!strcmp("Name", key) || !strcmp(entry.buf, key)) {
				sbuf_init(&name);
				sbuf_addstr(&name, strtok_r(rest, "=", &rest));
            }
            else if (!strcmp("Exec", key)) {
				sbuf_init(&exec);
				sbuf_addstr(&exec, strtok_r(rest, "=", &rest));
            }
            else if (!strcmp("Icon", key)) {
				sbuf_init(&icon);
                const char *str = strtok_r(rest, "=", &rest);
                get_menu_icon_path (&icon, str);
            }
            else if (!strcmp("NoDisplay", key)) {
                if (!strcmp(strtok_r(rest, "=", &rest), "true"))
                    display_ok = false;
            }
            else if (!strcmp("Terminal", key)) {
                if (!strcmp(strtok_r(rest, "=", &rest), "true"))
                    terminal_ok = true;
            }
            else if (!strcmp("ShowOnlyIn", key)) {
                show_item_ok = false;
                gchar *tokenizer = strtok_r(rest, "=", &rest);
                gchar **arr = g_strsplit_set (tokenizer, ";", -1);
                int n = g_strv_length(arr) - 1;
                for(int i=0; i<n; i++) {
                    if (!g_strcmp0(arr[i], "LXDE"))
                       show_item_ok = true;
                }            
                g_strfreev (arr);
            }
            else if (!strcmp("Type", key)) {
				sbuf_init(&type);
				sbuf_addstr(&type, strtok_r(rest, "=", &rest));
            }
            else if (!strcmp("Categories", key)) {
                gchar *tokenizer = strtok_r(rest, "=", &rest);
                gchar **arr = g_strsplit_set (tokenizer, ";", -1);
                int n = g_strv_length(arr) - 1;
                for(int i=0; i<n; i++) {
					unsigned long len = sglib_cstr_vector_size (dir_ctx->label);
					int m = 0;
					while(m<len) {
                       if (!g_strcmp0(arr[i], sglib_cstr_vector_at (dir_ctx->label, m))) {
				          sbuf_init(&category);
				          sbuf_addstr(&category, sglib_cstr_vector_at (dir_ctx->name, m));
                          break;
                       }
                       m++;
                    }
                }            
                g_strfreev (arr);
            }
        }
    }
    
    if(terminal_ok) {
		
	   char *tmp = strdup_or_null(exec.buf);
	   sbuf_init(&exec);
	   sbuf_concat(&exec, 2, " -e ", tmp);
	   free(tmp);
    }
    
    if(display_ok && show_item_ok && !strcmp(type.buf, "Application")) {
		
       rc = sglib_cstr_vector_push_back (app_ctx->icon, strdup_or_null(icon.buf));
       if(rc < 0) goto Finish_display_app;
       
       rc = sglib_cstr_vector_push_back (app_ctx->name, strdup_or_null(name.buf));
       if(rc < 0) goto Finish_display_app;
       
       rc = sglib_cstr_vector_push_back (app_ctx->exec, strdup_or_null(exec.buf));
       if(rc < 0) goto Finish_display_app;
       
       rc = sglib_cstr_vector_push_back (app_ctx->category, strdup_or_null(category.buf));
       if(rc < 0) goto Finish_display_app;
    } 

Finish_display_app:

    free(eol);
    sbuf_free(&name);
    sbuf_free(&exec);
    sbuf_free(&icon);
    sbuf_free(&type);
    sbuf_free(&category);
    sbuf_free(&entry);
    
    fclose (fp);
    return rc;
}

void get_menu_icon_path (sbuf_t *s, const char *name)
{
    GtkIconInfo *icon_info = NULL;
    GtkIconTheme *icon_theme;
        
    icon_theme = gtk_icon_theme_get_default ();

    if (name)
    {
        gchar *tmp_name = NULL;
        
        if (g_path_is_absolute (name)) {
			
			sbuf_addstr(s, name);
            return;
        }    

        tmp_name = strndup (name, strrchr (name, '.') - name);
        icon_info = gtk_icon_theme_lookup_icon (icon_theme, tmp_name, 24, GTK_ICON_LOOKUP_NO_SVG |
                                                                          GTK_ICON_LOOKUP_GENERIC_FALLBACK);
        g_free (tmp_name);

        if (!g_path_is_absolute(name) && !icon_info) {
            
            sbuf_t png, xpm, svg;
            
            sbuf_init(&png);            
            sbuf_init(&xpm);            
            sbuf_init(&svg);
            
            sbuf_concat(&png, 2, name, ".png");            
            sbuf_concat(&xpm, 2, name, ".xpm");            
            sbuf_concat(&svg, 2, name, ".svg");
            
            find_missing_icon (s, "/usr/share/icons/hicolor/24x24/apps", 4, name, png.buf, xpm.buf, svg.buf);
            if(*s->buf == 0) {
                find_missing_icon (s, "/usr/share/icons/hicolor/32x32/apps", 4, name, png.buf, xpm.buf, svg.buf);
                if(*s->buf == 0) {
                    find_missing_icon (s, "/usr/share/icons/hicolor/48x48/apps", 4, name, png.buf, xpm.buf, svg.buf);
                    if(*s->buf == 0) {
                        find_missing_icon (s, "/usr/share/icons/hicolor/256x256/apps", 4, name, png.buf, xpm.buf, svg.buf);
                        if(*s->buf == 0) {
                            find_missing_icon (s, "/usr/share/icons/hicolor/scalable/apps", 2, name, svg.buf);
                        }
                    }
                }
            }
            
            sbuf_free(&png);
            sbuf_free(&xpm);
            sbuf_free(&svg);
        }
    }
    
    if (!icon_info)
       icon_info = gtk_icon_theme_lookup_icon (icon_theme, "empty", 24, GTK_ICON_LOOKUP_NO_SVG);
        
    if(*s->buf == 0)
       sbuf_addstr(s, gtk_icon_info_get_filename (icon_info));
    
    gtk_icon_info_free (icon_info);
}

void find_missing_icon (sbuf_t *s, const char *path, int count, ...)
{
    va_list items;
    struct dirent *ent;
    
    DIR *dirp = opendir (path);
    while(ent = readdir(dirp))
    {
        int i=0;
        va_start(items, count);  // Requires the last fixed parameter (to get the address)
        while (i<count) {            
            if((fnmatch(va_arg(items, const gchar*), ent->d_name,0)) == 0) {
                sbuf_t tmp;
                sbuf_init(&tmp);
                sbuf_concat(&tmp, 3, path, "/", ent->d_name);   
                char *str = rindex(tmp.buf, '.');                                
                if(!strcmp(tmp.buf, ".svg")) {
                    sbuf_t icon_dir, new_icon;
                    sbuf_init(&icon_dir);
                    sbuf_init(&new_icon);
                    sbuf_concat(&icon_dir, 2, getenv("HOME"), "/.config/jwm-menu-cache");
                    mkdir(icon_dir.buf, 0777);
                    const char *basename = strtok( rindex(strdup_or_null(tmp.buf), '/'), "." );
                    sbuf_concat(&new_icon, 2, basename, ".png");
                    rsvg_convert(tmp.buf, new_icon.buf);
                    sbuf_addstr(s, new_icon.buf);
                    sbuf_free(&icon_dir);
                    sbuf_free(&new_icon);
                    sbuf_free(&tmp);
                    break;                    
                } else {                    
                    sbuf_addstr(s, tmp.buf);
                    sbuf_free(&tmp);
                    break;
                }
            }
            
            i++;
        }
        
        if(*s->buf != 0) break;        
    }
    
    va_end(items);
    closedir(dirp);
}

int display_dir (struct dir_context *dir_ctx, const char *path)
{
    int rc = 0;
    char buffer[MAX_SIZE] = {0};
    char *eol;
    sbuf_t name, icon, entry, type;
    FILE *fp = NULL;
    
    fp = fopen(path, "r");
    if(!fp) {
        exit(EXIT_FAILURE);
    }
    
    eol = malloc(MAX_SIZE * sizeof(char));  
    if(!eol) {
        printf("Memory failure\n");
        exit( EXIT_FAILURE );
    }
    
    while(eol[0] != '[') fgets(eol, MAX_SIZE, fp);

    sbuf_init(&entry);
    if(lang && !strcmp(lang, "en"))
        sbuf_addstr(&entry, "Name");
    else
        sbuf_concat(&entry, 4, "Name", "[", lang, "]");
        
    while(fgets(eol, MAX_SIZE, fp)>0) {
        eol[strcspn(eol, "\n")] = 0;
        if(eol[0] == '[') break;
        else if(eol[0] == '#' || isalpha(eol[0]) == 0) continue; 
        else {
            
            char *rest = eol;
            char *key = strtok_r(rest, "=", &rest);
            
            if (!strcmp(entry.buf, key)) {
				sbuf_init(&name);
				sbuf_addstr(&name, strtok_r(rest, "=", &rest));
            }
            else if (!strcmp("Icon", key)) {
				sbuf_init(&icon);
                const char *str = strtok_r(rest, "=", &rest);
                get_menu_icon_path (&icon, str);
            }
            else if (!strcmp("Type", key)) {
				sbuf_init(&type);
				sbuf_addstr(&type, strtok_r(rest, "=", &rest));
            }
        }
    }
   
    if(!strcmp(type.buf, "Directory")) {
		
       if(*icon.buf) {
          rc = sglib_cstr_vector_push_back (dir_ctx->icon, strdup_or_null(icon.buf));
          if(rc < 0) goto Finish_display_dir;
       }

       rc = sglib_cstr_vector_push_back (dir_ctx->name, strdup_or_null(name.buf));
       if(rc < 0) goto Finish_display_dir;
    }

Finish_display_dir:

    free(eol);
    sbuf_free(&name);
    sbuf_free(&icon);
    sbuf_free(&entry);
    sbuf_free(&type);
    
    fclose (fp);
    return rc;
}

gchar* parse_cmd (const gchar *exec)
{
    char *token, *u, *v;
    const char delim[2] = " ";
    gchar *filepath = NULL;
    
    u = malloc(64 * sizeof(char));  
    if(!u) {
        printf("Memory failure\n");
        exit( EXIT_FAILURE );
    }
    
    v = malloc(64 * sizeof(char));  
    if(!v) {
        printf("Memory failure\n");
        exit( EXIT_FAILURE );
    }
    
    g_return_val_if_fail(exec,"");

    GString *cmd = g_string_sized_new (64);
            
    for (;*exec; ++exec)
    {
        if (*exec == '%')
        {
            ++exec;
            switch (*exec)
            {
                case 'u':
                case 'U':
                case 'f':
                case 'F': break;
                case 'd':
                case 'D':
                case 'm':
                case 'n':
                case 'N':
                case 'v': break;
                default:
                    g_string_append_c (cmd, '%');
                    g_string_append_c (cmd, *exec);
                    break;
            }
        }
        else
            g_string_append_c (cmd, *exec);
    }
    
    v = g_strchomp (g_string_free (cmd, FALSE));
    strcpy (u, v);
    token  = strtok (v, delim);
    
    GString *str = g_string_new (u);
    free (u);
    free (v);
    return g_strchomp (g_string_free (str, FALSE));
}

int display_header_item (struct app_context *app_ctx, const char *filename)
{
    int rc = 0;
    node_t item;
    char buffer[MAX_SIZE] = {0};
    char *eol;
    sbuf_t name, exec, icon, entry, path;
    
    FILE *fp = NULL;
  
    sbuf_init(&path);
    sbuf_concat(&path, 3, "/usr/share/applications/", filename, ".desktop");
    
    if(access(path.buf, F_OK) != 0) exit( EXIT_FAILURE );
    
    fp = fopen(path.buf, "r");
    if(!fp) {
        exit(EXIT_FAILURE);
    }
    
    sbuf_free(&path);
    
    eol = malloc(MAX_SIZE * sizeof(char));  
    if(!eol) {
        printf("Memory failure\n");
        exit( EXIT_FAILURE );
    }
    
    while(eol[0] != '[') fgets(eol, MAX_SIZE, fp);
    
    sbuf_init(&entry);
    if(lang && !strcmp(lang, "en"))
        sbuf_addstr(&entry, "Name");
    else
        sbuf_concat(&entry, 4, "Name", "[", lang, "]", NULL);
        
    while(fgets(eol, MAX_SIZE, fp)>0) {
        eol[strcspn(eol, "\n")] = 0;
        if(eol[0] == '[') break;
        else if(eol[0] == '#' || isalpha(eol[0]) == 0) continue; 
        else {
            
            char *rest = eol;
            char *key = strtok_r(rest, "=", &rest);
            
            if (!strcmp("Name", key) || !strcmp(entry.buf, key)) {
				sbuf_init(&name);
				sbuf_addstr(&name, strtok_r(rest, "=", &rest));
            }
            else if (!strcmp("Exec", key)) {
				sbuf_init(&exec);
				sbuf_addstr(&exec, strtok_r(rest, "=", &rest));
            }
            else if (!strcmp("Icon", key)) {
				sbuf_init(&icon);
                const char *str = strtok_r(rest, "=", &rest);
                get_menu_icon_path (&icon, str);
            }
        }
    }
    
    if(*exec.buf && *name.buf) {
/*
 unsigned long len = sglib_cstr_vector_size (app_ctx->icon);
 printf("len=%d\n", len);
 sglib_cstr_vector_set (app_ctx->icon, NULL, len - 1);
 sglib_cstr_vector_set (app_ctx->name, NULL, len - 1);
 sglib_cstr_vector_set (app_ctx->exec, NULL, len - 1);
*/
       if(*icon.buf) {
          rc = sglib_cstr_vector_push_back (app_ctx->icon, strdup_or_null(icon.buf));
          if(rc < 0) goto Finish_display_header_item;
       }

       rc = sglib_cstr_vector_push_back (app_ctx->name, strdup_or_null(name.buf));
       if(rc < 0) goto Finish_display_header_item;

       rc = sglib_cstr_vector_push_back (app_ctx->exec, strdup_or_null(exec.buf));
       if(rc < 0) goto Finish_display_header_item;
       
       rc = sglib_cstr_vector_push_back (app_ctx->category, "menu_header");
       if(rc < 0) goto Finish_display_header_item;
    }   

Finish_display_header_item:
       
    free(eol);
    sbuf_free(&name);
    sbuf_free(&exec);
    sbuf_free(&icon);
    sbuf_free(&entry);
    
    fclose (fp);

    return rc;
}

// buf must be freed after usage
int get_exec_command_from_filename (const char *filename, char **buf)
{
    node_t item;
    char buffer[MAX_SIZE] = {0};
    char *eol;
    FILE *fp = NULL;
    sbuf_t path;
        
    *buf = (char*)malloc(sizeof(char) * 64);
    if (!*buf) {
        fprintf(stderr, "Memory allocation failure\n");
        exit (EXIT_FAILURE);
    }
    
    sbuf_init(&path);    
    sbuf_concat(&path, 3, "/usr/share/applications/", filename, ".desktop");
    
    if(access(path.buf, F_OK) != 0) return 1; 
    
    fp = fopen(path.buf, "r");
    if(!fp) {
        exit(EXIT_FAILURE);
    }
    
    sbuf_free(&path);
    
    eol = malloc(MAX_SIZE * sizeof(char));  
    if(!eol) {
        printf("Memory failure\n");
        exit( EXIT_FAILURE );
    }
        
    while(fgets(eol, MAX_SIZE, fp)>0) {
        eol[strcspn(eol, "\n")] = 0;
        if(eol[0] == '[') continue;
        else if(eol[0] == '#' || isalpha(eol[0]) == 0) continue; 
        else {
            
            char *rest = eol;
            char *key = strtok_r(rest, "=", &rest);
            
            if (!strcmp("Exec", key))        
                sprintf(*buf, "%s", strtok_r(rest, "=", &rest));
        }
    }

    free(eol);
    
    fclose (fp);
    
    return 0;
}

void kill_all_instances()
{
	DIR *dirp = NULL;
	struct dirent *dir = NULL;
    const char *processes[ARRAY_DIM] = { "dxmenu-cache" };
    pid_t pid;
    char pidbuf[10];
    
    pid = getpid();
    pidbuf[9] = 0;
    itoa10_safe(pid, pidbuf);
	
	dirp=opendir("/proc");
	if(!dirp) {
		printf("\n\tERROR: unable to open /proc for read\n\n");
		exit(EXIT_FAILURE);
	}
	
	while((dir=readdir(dirp)) != NULL ) {

        int rc = 0;
	    struct stat sb;
        char *proc_cmd = NULL;
        bool is_uint_ok = true;
        char pth[MAX_SIZE] = {0};
	    
	    for(unsigned i=0; i<strlen(dir->d_name); i++) {
		   if(!isdigit(dir->d_name[i])) is_uint_ok = false;
	    }
        
        if(!is_uint_ok) continue;
    
        strncpy_t(pth, sizeof(pth), "/proc/", strlen("/proc/"));
        strcat(pth, dir->d_name);
        if (stat(pth, &sb) == -1) {
           perror("stat");
           // do not exit, go ahead
           continue;
        }
        
        if(!strcmp(dir->d_name, pidbuf)) continue;
          
        if(!lstat(dir->d_name, &sb)) continue;
 
        /* Only consider directories */
        if(!S_ISDIR(sb.st_mode)) continue;
        
		rc = pstat_get_binary_path_from_string(&proc_cmd, MAX_SIZE, dir->d_name);
 
        for(unsigned int i=0; i<ARRAY_DIM; i++) {		                    
           if(proc_cmd && !strcmp(rindex(proc_cmd, '/') + 1, processes[i])) {
              if( (kill(atoi(dir->d_name), SIGTERM)) != 0 ) kill(atoi(dir->d_name), SIGKILL);
           }
        }

	    free(proc_cmd);
	}
	
    closedir(dirp);
}

/*
 *    SBUF
 * 
 */ 

void sbuf_init(sbuf_t *s)
{
	s->buf = (char*)malloc(1);
	s->buf[0] = '\0';
	s->bufsiz = 1;
	s->len = 0;
}

void sbuf_addch(sbuf_t *s, char ch)
{
	if (s->bufsiz <= s->len + 1) {
		s->bufsiz = s->bufsiz * 2 + 16;
		s->buf = (char*)realloc(s->buf, s->bufsiz);
	}
	s->buf[s->len++] = ch;
	s->buf[s->len] = 0;
}

void sbuf_addstr(sbuf_t *s, const char *data)
{
	int len;

	if (!data)
		return;
	len = strlen(data);
	if (s->bufsiz <= s->len + len + 1) {
		s->bufsiz = s->bufsiz + len;
		s->buf = realloc(s->buf, s->bufsiz);
	}
	memcpy(s->buf + s->len, data, len);
	s->len += len;
	s->buf[s->len] = 0;
}

/**
 * \brief Variadic function
 */
void sbuf_concat(sbuf_t *s, int count, ...)
{
    va_list items;
	
    va_start(items, count);  /* Requires the last fixed parameter (to get the address) */
    
    while(count--)
		sbuf_addstr(s, va_arg(items, const char*));
	
    va_end(items);
}

void sbuf_free(sbuf_t *s)
{
	if(!s->buf) return;
	free(s->buf);
	s->buf = NULL;
}

